# module level doc-string
__doc__ = """
General logging utility module for KDS

@author: Manoj Bonam
Created on Thu Mar 11 14:57:55 2021

DEBUG: Detailed information, typically of interest only when diagnosing problems.
INFO: Confirmation that things are working as expected.
WARNING: An indication that something unexpected happened, or indicative of some problem in the near future (e.g. ‘disk space low’). The software is still working as expected.
ERROR: Due to a more serious problem, the software has not been able to perform some function.
CRITICAL: A serious error, indicating that the program itself may be unable to continue running.

################################################# INSTRUCTIONS #########################################################
# In your program always import the entire kdsutil module first, to properly set the local variables.                  #
# from kdslib import kdsutil                                                                                           #
# The main class can be instantiated like this: loggingutil = kdsutil.Generate_Logger(**kwargs)                        #
# The class variables can be accessed in this format: loggingutil.logger.info('Hello World')                           #
# The class methods can be accessed in this format: loggingutil.log_info('Hello World')                                #
########################################################################################################################

Library: https://docs.python.org/3/library/logging.html

"""

import os, logging, yaml, sys, glob, json, traceback, shutil, math
from pathlib import Path
import datetime, time
import teradata, teradatasql #teradatasqlalchemy
import urllib.parse as urllib
from urllib.parse import quote_plus as quote_plus
import pyodbc, pandas as pd
from functools import wraps
from sqlalchemy import create_engine

# Import smtplib for the actual sending function
import smtplib
# Import the email modules we'll need
from email.message import EmailMessage

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


#================================================================
# Check if environment variable is set
#================================================================
try:
    if os.name == 'nt': # This is for windows
        _env = os.environ['kds-python-config']
    else: # This is for mac
        _env = os.environ['kds_python_config']
    print("User's Environment variable:" + str(_env))
except:
    print("Configuration file _env variable is not set.")
    print("set kds-python-config variable")
    sys.exit(1)


    
#===============================================================
# If environment variable is set then read the config yaml file
#===============================================================
with open(_env,"r") as yamlConfig:
    cfg = yaml.safe_load(yamlConfig)

#################################################################
# Generate a logger 
#################################################################
class Generate_Logger:  
    """
    When instantiated, creates a logging object.
    
    """    
    #Main constructor of the Generate_Logger class.
    def __init__(self,*args, **kwargs):
        self.logger = logging.getLogger(__name__)
        
        if kwargs.get('jobFrequency') is None:
            self.timeStampObj = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        elif kwargs.get('jobFrequency').lower().strip() == 'high':
            self.timeStampObj = datetime.datetime.now().strftime('%Y%m%d')
        else:
            self.timeStampObj = datetime.datetime.now().strftime('%Y%m%d%H%M%S') 
            
        self.programName = kwargs.get('programName')
        
        if kwargs.get('nestedDirName') is None:
            self.fileName = cfg.get('KDS_LOG').get('log_loc') + self.timeStampObj + "_" + self.programName + ".log"
        else:
            self.fileName = Path(cfg.get('KDS_LOG').get('log_loc') + '\\' + kwargs.get('nestedDirName') + '\\' + self.timeStampObj + "_" + self.programName + ".log")
            
        self.appname = kwargs.get('appname')
        self.loglevel= cfg.get(self.appname).get('log_level')
        
        if self.loglevel.lower() == "info":
            self.logger.setLevel(logging.INFO)
        elif self.loglevel.lower() == "warning":
            self.logger.setLevel(logging.WARNING)
        elif self.loglevel.lower() == "debug":
            self.logger.setLevel(logging.DEBUG)
        elif self.loglevel.lower() == "error":
            self.logger.setLevel(logging.ERROR)

        if self.logger.handlers:
            self.logger.handlers = []
   
        self.file_handler = logging.FileHandler(self.fileName)
        self.formatter = logging.Formatter('%(asctime)s: %(levelname)s: %(name)s: %(message)s')

        self.file_handler.setFormatter(self.formatter)

        self.stream_handler = logging.StreamHandler()
        self.stream_handler.setFormatter(self.formatter)

        self.logger.addHandler(self.file_handler)
        self.logger.addHandler(self.stream_handler)
        
    # Class methods that can be accessed using the class instance.
    def log_info(self, msg):
       self.logger.info(msg)
       
    def log_warning(self, msg):
        self.logger.warning(msg)        
    
    def log_error(self, msg):
        self.logger.error(msg)
    
    def log_debug(self, msg):
        self.logger.debug(msg)  
     
    # This method cleans up the logging object/handlers and exits the log file.
    def delete(self):
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.flush()
            handler.close()
            self.logger.removeHandler(handler)   


#################################################################
# Templates for database connections
# Depending on the type of connection, proper key word arguments 
# must be passed to get the connection variable.
#################################################################

def get_kds_teradata_conn(**kwargs):
    
    trycntr = 1
    stoptrying = False
    
    while not stoptrying:

        try:   
            udaExec = teradata.UdaExec (appName='kds', version="2", configureLogging=False, logConsole = False)
            conn = udaExec.connect(DSN=kwargs.get('ODBC_name'),method="odbc")
            stoptrying = True
        except Exception as e:            
            if trycntr >= 3:
                print('Unable to connect to EDW: ' + str(e))
                conn = e
                stoptrying = True
                
            trycntr += 1
            time.sleep(3)

    return conn


def get_sqlalchemy_teradata_engine(**kwargs):

    if  kwargs.get('username') is not None:
        user_ = kwargs.get('username')
    elif kwargs.get('userid') is not None:
        user_ = kwargs.get('userid')
    elif kwargs.get('user') is not None:
        user_ = kwargs.get('user')
        
        
    if os.name == 'posix': # This is a mac
        edw_engine = create_engine("teradatasql://{host}/?logmech=LDAP&username={username}&password={password}"
                                   .format(host = quote_plus(kwargs.get('server')),
                                           username = quote_plus(user_),
                                           password = quote_plus(kwargs.get('password')))).connect()
        
    else:
        params_str = urllib.quote("""DRIVER={driver};DBCNAME={host};UID={username};PWD={password};authentication=LDAP;"""
                           .format(driver = quote_plus(kwargs.get('driver')),
                                   host = quote_plus(kwargs.get('server')),
                                   username = quote_plus(user_),
                                   password = quote_plus(kwargs.get('password'))))  
        
        conn_str = 'mssql+pyodbc:///?odbc_connect={params_str}'.format(params_str=params_str)
        
        edw_engine = create_engine(conn_str, fast_executemany=True)        

    return edw_engine


def get_pyodbc_windowsauth_conn(**kwargs):
   
    params_str = urllib.quote("""Driver={driver};Server={server};Database={database};Trusted_Connection=yes;""".format(driver=kwargs.get('driver'),
                                                   server=kwargs.get('server'),
                                                   database=kwargs.get('database')))
    conn_str = 'mssql+pyodbc:///?odbc_connect={params_str}'.format(params_str=params_str)    
    
    return conn_str


def get_pyodbc_windowsauth_sa_engine(**kwargs):
   
    params_str = urllib.quote("""Driver={driver};Server={server};Database={database};Trusted_Connection=yes;""".format(driver=kwargs.get('driver'),
                                                   server=kwargs.get('server'),
                                                   database=kwargs.get('database')))
    conn_str = 'mssql+pyodbc:///?odbc_connect={params_str}'.format(params_str=params_str)
    
    sqlsaengine = create_engine(conn_str, fast_executemany=True)
    
    return sqlsaengine


def get_pyodbc_windowsauth_cnxn(**kwargs):

    params_str = ("Driver={driver};Server={server};Database={database};Trusted_Connection={trusted_connection};").format(
        driver=kwargs.get('driver'),
        server=kwargs.get('server'),
        database=kwargs.get('database'),
        trusted_connection="yes")
        
    conn = pyodbc.connect(params_str)   
    
    return conn    


def get_pyodbc_credentials_conn(**kwargs):
    
    if  kwargs.get('username') == None:
        user_ = kwargs.get('user')
    else:
        user_ = kwargs.get('username')
    
    params_str = ("Driver={driver};Server={server};Database={database};Uid={user};Pwd={password};").format(
                                            driver=kwargs.get('driver'),
                                            server=kwargs.get('server'),
                                            database=kwargs.get('database'),
                                            user=user_,
                                            password=kwargs.get('password'))
    
    conn = pyodbc.connect(params_str)
    
    return conn


def get_pyodbc_credentials_sa_engine(**kwargs):
    
    if  kwargs.get('username') == None:
        user_ = kwargs.get('user')
    else:
        user_ = kwargs.get('username')
    
    params_str = urllib.quote("""Driver={driver};Server={server};Database={database};Uid={user};Pwd={password};""".format(
                                                   driver=kwargs.get('driver'),
                                                   server=kwargs.get('server'),
                                                   database=kwargs.get('database'),
                                                   user=user_,
                                                   password=kwargs.get('password')
                                                   ))
    conn_str = 'mssql+pyodbc:///?odbc_connect={params_str}'.format(params_str=params_str)
   
  
    
    sqlACEngine = create_engine(conn_str, fast_executemany=True)
    
    return sqlACEngine


def get_pyodbc_azure_ad_auth_conn(**kwargs):
    params_str = ("Driver={driver};Server={server};Database={database};Authentication={auth};").format(
        driver=kwargs.get('driver'),
        server=kwargs.get('server'),
        database=kwargs.get('database'),
        auth=kwargs.get('auth'))
    
    conn = pyodbc.connect(params_str)
    
    return conn

    
def get_pyodbc_azure_credentials_conn(**kwargs):
    
    if  kwargs.get('username') == None:
        user_ = kwargs.get('user')
    else:
        user_ = kwargs.get('username')
        
    params_str = ("Driver={driver};Server={server};Database={database};Uid={username};Pwd={password};").format(
        driver=kwargs.get('driver'),
        server=kwargs.get('server'),
        database=kwargs.get('database'),
        username=user_,
        password=kwargs.get('password'))
    
    conn = pyodbc.connect(params_str)
    
    return conn   


def get_sqlalchemy_sql_credentials_engine(**kwargs):

    if  kwargs.get('username') is not None:
        user_ = kwargs.get('username')
    elif kwargs.get('userid') is not None:
        user_ = kwargs.get('userid')
    elif kwargs.get('user') is not None:
        user_ = kwargs.get('user')
        
    if os.name == 'posix': # This is a mac
        sql_engine = create_engine("mssql+pymssql://{username}:{password}@{host}/{database}".format(host=quote_plus(kwargs.get('server')),
                                                                                                    database=quote_plus(kwargs.get('database')),
                                                                                                    username=quote_plus(user_),
                                                                                                    password=quote_plus(kwargs.get('password')))).connect()            
    else:
        params_str = urllib.quote("""Driver={driver};Server={server};Database={database};Uid={user};Pwd={password};""".format(
                                                       driver=quote_plus(kwargs.get('driver')),
                                                       server=quote_plus(kwargs.get('server')),
                                                       database=quote_plus(kwargs.get('database')),
                                                       user=quote_plus(user_),
                                                       password=quote_plus(kwargs.get('password'))))
               
        conn_str = 'mssql+pyodbc:///?odbc_connect={params_str}'.format(params_str=params_str)      
        
        sql_engine = create_engine(conn_str, fast_executemany=True)
       
    return sql_engine

    
def senderroremail(**kwargs):
    
    programName_ = kwargs.get('programName')

    msg = EmailMessage()
    msg.set_content("There was an error running " + programName_ + os.linesep + kwargs.get('errStackTrace'))
    msg['Subject'] = 'Error executing ' + programName_
    msg['From'] = cfg.get('EMAIL').get('reply')
    msg['To'] = cfg.get(kwargs.get('appName')).get('support_email')
    
    # Send the message via our own SMTP server.
    if  cfg.get('EMAIL').get('SMTP') is not None:
        s = smtplib.SMTP(cfg.get('EMAIL').get('SMTP'))
    else:
        s = smtplib.SMTP(cfg.get('EMAIL').get('smtp'))

    s.send_message(msg)
    s.quit()
    
    return

    
def emaillogfile(**kwargs):
    
    programName_ = kwargs.get('programName')
    fileName_ = kwargs.get('fileName')

    msg = EmailMessage()
    msg.set_content("Please find the attached file for  " + programName_ + ".py's latest log.")
    msg['Subject'] = 'Log for ' + programName_ + '.py'
    msg['From'] = cfg.get('EMAIL').get('reply')
    msg['To'] = cfg.get(kwargs.get('appName')).get('support_email')
    try:
        msg.add_attachment(open(fileName_, "r").read(), filename=fileName_.split('\\')[-1])
    except Exception as e:
        msg.set_content("The KDSutil was unable to not find the log file for  " + programName_ + ".py's latest run.")
        
    # Send the message via our own SMTP server.
    if  cfg.get('EMAIL').get('SMTP') is not None:
        s = smtplib.SMTP(cfg.get('EMAIL').get('SMTP'))
    else:
        s = smtplib.SMTP(cfg.get('EMAIL').get('smtp'))
        
    s.send_message(msg)
    s.quit()
    
    return


def sendemailnotification(**kwargs):
    
    programName_ = kwargs.get('programName')

    msg = EmailMessage()
	
    if  kwargs.get('emailContent') is not None:
        emailContent_ = kwargs.get('emailContent')
    else:
        emailContent_ = 'This is the requested email for '	+ programName_

    msg['From'] = cfg.get('EMAIL').get('reply')

    if  kwargs.get('recipient') is not None:
        msg['To'] = kwargs.get('recipient') 
    else:
        msg['To'] = cfg.get(kwargs.get('appName')).get('support_email')
        
    msg.set_content(emailContent_)
    
    if  kwargs.get('subject') is not None:
        msg['Subject'] = kwargs.get('subject')
    else:
        msg['Subject'] = 'Email Notification For ' + programName_ + '.py'
	
    # Send the message via our own SMTP server.
    if  cfg.get('EMAIL').get('SMTP') is not None:
        s = smtplib.SMTP(cfg.get('EMAIL').get('SMTP'))
    else:
        s = smtplib.SMTP(cfg.get('EMAIL').get('smtp'))
        
    s.send_message(msg)
    s.quit()
    
    return


def sendmimeemailmessage(**kwargs):
    
    programName_ = kwargs.get('programName')

    #Create message container.
    msgRoot = MIMEMultipart('related')
	
    if  kwargs.get('subject') is not None:
        msgRoot['Subject'] = kwargs.get('subject')
    else:
        msgRoot['Subject'] = 'Email Notification For ' + programName_ + '.py'

	
    if  kwargs.get('emailContent') is not None:
        HTMLemailContent_ = kwargs.get('emailContent')
    else:
        HTMLemailContent_ = 'This is the requested email for '	+ programName_
		
    msgHtml = MIMEText(HTMLemailContent_, 'html')
    msgRoot.attach(msgHtml)

    if  kwargs.get('recipient') is not None:
        msgRoot['To']  = notification_email_to= kwargs.get('recipient') 
    else:
        msgRoot['To']  = notification_email_to = cfg.get(kwargs.get('appName')).get('support_email')
	
    # Send the message via our own SMTP server.
    if  cfg.get('EMAIL').get('SMTP') is not None:
        s = smtplib.SMTP(cfg.get('EMAIL').get('SMTP'))
    else:
        s = smtplib.SMTP(cfg.get('EMAIL').get('smtp'))
        
    s.sendmail(cfg.get('EMAIL').get('reply'), notification_email_to, msgRoot.as_string())
    s.quit()
    
    return
    

def read_parquet_files(path, match_pattern=None):    
    listOfParquetFiles = glob.glob(path + "/*.parquet") if match_pattern is None else glob.glob(path + f"/*{match_pattern}*.parquet")
    df = [pd.read_parquet(f) for f in listOfParquetFiles]
    df_all = pd.concat(df,ignore_index=True)
    df_all.reset_index(inplace=True, drop=True)

    return df_all


def create_folder(folder, DeleteIfExists):
    if DeleteIfExists:
        if Path(folder).is_dir():
            shutil.rmtree(folder)
            os.makedirs(folder)
    else:
        if not os.path.exists(folder):
            os.makedirs(folder)
        
    return


def generate_metadata(*args):
    
    file = args[0]
    script = args[1]
    version = args[2]
    source = args[3]
    source_sql = args[4]
    description = args[5]
    df = args[6]    

    file = file.replace('\\','/')
    script = script.replace('\\','/')
    file_name = file.rsplit('/',1)[1]
    path = file.replace(file_name,'')
    path = path[:-1] if path[-1]=='/' else path
    json_str = '''{
    
                "file": "''' + file_name + '''",
                "path": "''' + path + '''",
                "version": "''' + version + '''",
                "source": "''' + source + '''",
                "source_sql": ''' + json.dumps(source_sql) + ''',
                "description": "''' + description + '''",
                "created on": "''' + datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S") + '''",
                "created by": "''' + script + '''",
                "data_structure": [ '''
    for ind, column in enumerate(df.columns):
        json_str = json_str + '{' + f' "position": "{ind+1}", "column": "{column}", "dtype": "' + df.dtypes[column].name + '"},'
    json_str = json_str[:-1] if json_str[-1]==',' else json_str
    json_str = json_str + ''']
    
    }'''
                                   
    json_file = file.replace('.parquet','') + '.json'
    with open(json_file, 'w', encoding='utf-8') as f:
        json.dump(json.loads(json_str), f, ensure_ascii=False, indent=4)

    return

#This decorator function adds timer to your functions
# can be called by adding the below at the top of your function
# @kdsutil.fn_execution_timer(logutilobj)
def fn_execution_timer(_loggerobj):
    def decorator(orig_func):
        @wraps(orig_func)
        def wrapper(*args, **kwargs):
            t1 = time.time()
            result = orig_func(*args, **kwargs)
            t2 = time.time() - t1
            print('{} ran in: {} seconds'.format(orig_func.__name__, t2))
            _loggerobj.log_info('{} ran in: {} seconds'.format(orig_func.__name__, t2))
            
            return result    
        return wrapper
    return decorator

#Pass in the Database name, Table name  and your dataframe to load it to the table eg: kdsutil.load_df_to_edw('EDW_PRD_SB_KDS','MX_TABLE',df)
#An optional kwarg is available in case it's not a kill and fill table eg: kdsutil.load_df_to_edw('EDW_PRD_SB_KDS','MX_TABLE',df, purge='N')

def load_df_to_edw(*args, **kwargs):    
    db  = args[0]
    table = args[1]
    df = args[2]

    if  kwargs.get('purge') is not None:
        if kwargs.get('purge').upper() == 'N' or  kwargs.get('purge').upper() == 'NO':
            purge = False
    else:
        purge = True
        
    if  kwargs.get('batchsize') is not None:
        nRowsPerBatch = int(kwargs.get('batchsize'))
    else:
        numcols = df.shape[1]
        nRowsPerBatch = 2500 if numcols < 50 else 500

    df.drop_duplicates(inplace=True)    
    df_lst = [tuple(x) for x in df.values]
    nRows = len(df)    
    nBatchCount = math.ceil(nRows/nRowsPerBatch)
    sTableName = db + '.' + table
    
    start = 0
    finish = nRowsPerBatch
    
    conn_edw = get_kds_teradata_conn(**cfg.get('edw'))
    
    if purge:   
        del_str = f"""DELETE FROM {sTableName};"""
        cursor = conn_edw.cursor()
        cursor.execute(del_str)            
    
    qstr = '?,'
    itr = 1
    
    for q in range(numcols-1):  
        if itr == numcols-1:
            qstr += '?'
            break
        qstr += '?,'     
        itr += 1
            
    with conn_edw as session:
        for batch in range(0, nBatchCount):
            session.executemany(f"""INSERT INTO {sTableName} VALUES ({qstr})"""
                                , df_lst[start:finish], batch=True)
            start = start + nRowsPerBatch
            finish = finish + nRowsPerBatch
            
    return 
 
#################################################################
# Main function equivalent. 
# Only gets executed when this file is run independently  
#################################################################
if __name__ == '__main__':          
    
   _filename = sys.argv[0].split(os.path.sep)[-1].split('.')[0]
   print(_filename)
   
   loggerdetails= {'appname':'KDS_UTIL'}
   #Create a logging instance of the Generate_Logger class.
   logutilobj = Generate_Logger(**loggerdetails)
   
   try:
       #Invoke the log_info class method for the logutilobj.
       logutilobj.log_info('Main function test begin.')
       result = 10/0
    
   except:
       err = sys.exc_info()[0]
       #Invoke the log_error class method for the logutilobj.
       logutilobj.log_error('Tried to divide by zero')
       errStackTrace = traceback.format_exc(limit=None, chain=True)
       #Invoke the log_error class method for the logutilobj.
       logutilobj.log_error(errStackTrace)
       logutilobj.delete() 
       print('completed the except section')


# module level doc-string
__doc__ = """
Generic functional utility module for KDS

@author: Manoj Bonam
Created on Fri Jul 30 15:29:55 2021

################################################# INSTRUCTIONS #########################################################
# In your program always import the entire kdsutil module first, to properly set the local variables.                  #
# from kdslib import kdsfunctionalutil                                                                                 #
########################################################################################################################

"""


import os,yaml,sys
import msal,requests,json

#================================================================
# Check if environment variable is set
#================================================================
try:
    env = os.environ['kds-python-config']
    print("User's Environment variable:" + str(env))
except:
    print("Configuration file env variable is not set.")
    print("set kds-python-config variable")
    sys.exit(1)
    
#===============================================================
# If environment variable is set then read the config yaml file
#===============================================================
with open(env,"r") as yamlConfig:
    cfg = yaml.safe_load(yamlConfig)


"""
   The refreshPBIdataset function is used to refresh a PBI dataset. The workspaceID and datasetID have to be packed and passed
   as a dict to this function. 
   For a successful refresh, these(one or both, based on test vs prod) accounts have to be added as Members to your workspace/dataset
   
   KDS_Data_Scheduler: '2d031093-51be-4a6e-8608-ef7f19a1643e'
   KDS_Data_Scheduler_NonProd: 'd302df67-f061-4a65-a2bc-07eda5e89747'

"""

def refreshPBIdataset(**kwargs):
    
    workspaceID_ = kwargs.get('WORKSPACEID')
    datasetID_ = kwargs.get('DATASETID')
    authority_url = cfg.get('PBI_OnDemand_Refresh').get('authority_url')
    resource_url = cfg.get('PBI_OnDemand_Refresh').get('resource_url')
    client_id = cfg.get('PBI_OnDemand_Refresh').get('client_id')
    clientsecret_ = cfg.get('PBI_OnDemand_Refresh').get('client_secret')
    
    app = msal.ConfidentialClientApplication(client_id,  #confidential client
                                             authority=authority_url,
                                             client_credential=clientsecret_)
    

    result = app.acquire_token_for_client(scopes=resource_url)
    access_token = result['access_token']   
    
    #refresh_url = 'https://api.powerbi.com/v1.0/myorg/groups/WORKSPACEID/datasets/DATASETID/refreshes'
    refresh_url = f'''https://api.powerbi.com/v1.0/myorg/groups/{workspaceID_}/datasets/{datasetID_}/refreshes'''
    header = {'Authorization': f'Bearer {access_token}'}
    r = requests.post(url=refresh_url, headers=header)
    return r.raise_for_status()


"""
   The refreshInformaticaTaskFlows function is used to trigger task flows jobs in IICS. The taskflowID has to be packed and passed
   as a dict to this function. 
   
   response = kdsfunctionalutil.refreshInformaticaTaskFlows(**{'taskflowID':'s_SSRS_Taskflow_Daily_Weekly'})
   
"""
def refreshInformaticaTaskFlows(**kwargs):
    
    session_url = cfg.get('Informatica_OnDemand_Refresh').get('Session_URL')
    trigger_url = cfg.get('Informatica_OnDemand_Refresh').get('url')
    user = cfg.get('Informatica_OnDemand_Refresh').get('User')
    password = cfg.get('Informatica_OnDemand_Refresh').get('Password')
    taskflowID = kwargs.get('taskflowID')

    #Get ICSESSIONID which will be used to trigger taskflow jobs in IICS
    payload = json.dumps({"username": user,
                          "password": password})
    
    headers = {'Accept': 'application/json',
               'Content-Type': 'application/json'}
    
    login_response = requests.request("POST", session_url, headers=headers, data=payload)  
    output = json.loads(login_response.text)
    inputkey = output["icSessionId"] ##icsessionId retrieved and assigned to variable
  
    #Trigger taskflow jobs in Informatica, we will use ICSESSIONID retrieved above along with other parameters
    headers = {'Content-Type': 'application/json', 
             'IDS-SESSION-ID': inputkey}
    
    finaltaskflow_url = trigger_url + taskflowID  #concatenate API with taskflow ID generated by Informatica
    response = requests.request("POST", finaltaskflow_url, headers=headers)    
                 
    return response.raise_for_status()
    

#################################################################
# Main function equivalent. 
# Only gets executed when this file is run independently  
#################################################################
if __name__ == '__main__': 
    #Try refresh Samuel Travis's test PBI report.    
    response = refreshPBIdataset(**{'WORKSPACEID' : '197e3ddf-5c6b-4779-a1c5-2233cb9d4fc4', 
                                      'DATASETID' : 'b85b1e65-7c4e-4132-b092-33b73ec21399'})
   
    print('refreshPBIdataset response:' + str(response))
    
    response = refreshInformaticaTaskFlows(**{'taskflowID':'s_SSRS_Taskflow_Daily_Weekly'})    
    if response is None or '202' in response:
        print('IICS refresh attempt was successful.')
    else:
        print('IICS refresh failed.' + str(response))
        raise    
        


# module level doc-string
__doc__ = """
General data request utility module for KDS

@author: Himanshu Gadgil
Created on Fri Jun 24 09:47:44 2022

# Functionality: Returns a dataset (in a pandas dataframe format) by combining #
#                parquet files in the KDS datalake for various classifications #
#                & datasets.                                                   #

"""

import os, yaml, sys, traceback
import datetime
import pandas as pd
import random
import inspect, socket
from kdslib import kdsutil

#================================================================
# Check if environment variable is set
#================================================================
try:
    if os.name == 'posix': # This is a mac
        v_env = os.environ['kds_python_config']
    else:
        v_env = os.environ['kds-python-config']
except:
    print("Configuration file env variable is not set.")
    print("set kds-python-config variable")
    sys.exit(1)
    
#===============================================================
# If environment variable is set then read the config yaml file
#===============================================================
with open(v_env,"r") as yamlConfig:
    cfg = yaml.safe_load(yamlConfig)

#################################################################
# Get the current file name
#################################################################
programName = os.path.basename(__file__)
appConfigName = "KDS_LOG"
ExecMode = cfg.get('PYTHON_EXEC_MODE').get('INTERACTIVE')

kds_data_root = cfg.get('KDS_DATA').get('data_lake_root')
kds_data_de = cfg.get('KDS_DATA').get('data_eng')
kds_data_ds = cfg.get('KDS_DATA').get('data_sci')
kds_de_data_root = kds_data_root + '/' + kds_data_de
kds_data_wrapper = kds_de_data_root + '/DATA_WRAPPER/DataWrapper_DataSets.json'
kds_data_wrapper_log_dir = kds_de_data_root + '/DATA_WRAPPER/EXEC_LOG/'
df_dl_wrapper = pd.read_json(kds_data_wrapper) 

#################################################################
# Generate a logger to log info
#################################################################
logutilobj = kdsutil.Generate_Logger(appname=appConfigName, programName=programName)

def get_data(data_classification_, data_set_list_):
    
    #separate the main and other datasets into individual dataframes so we can process them
    df_dl_wrapper_data = df_dl_wrapper[df_dl_wrapper['DataClassification'].str.lower() == data_classification_.lower()]
    df_dl_wrapper_main = df_dl_wrapper_data[df_dl_wrapper_data['IsMain'] == 1]
    df_dl_wrapper_other= df_dl_wrapper_data[df_dl_wrapper_data['IsMain'] == 0]

    #Get the main data parquet location and replace the constant values with config values
    data_file = df_dl_wrapper_main['ParquetFileAndLocation'].iloc[0]
    data_file = data_file.replace(r'{KDS_DL_ROOT}',kds_data_root)
    if (r'{KDS_DE_DATA}' in data_file):
        data_file = data_file.replace(r'{KDS_DE_DATA}',kds_data_de)
    if (r'{KDS_DS_DATA}' in data_file):
        data_file = data_file.replace(r'{KDS_DS_DATA}',kds_data_ds)
    #Also get the columns that we need to return from the main dataframe.
    main_columns = df_dl_wrapper_main['ReturnColumn'].iloc[0]
    main_dataset_name = df_dl_wrapper_main['DataSet'].iloc[0]
    #Read the main data parquet into a dataframe
    main_df = pd.read_parquet(data_file)
    #change all the column names to lowercase so we have conformity
    main_df.columns = main_df.columns.str.lower()
    #Get the last modified datetime so we can log it
    file_TimeStamp = datetime.datetime.fromtimestamp(os.stat(data_file).st_mtime).strftime("%m/%d/%Y %H:%M:%S")
    row_count = len(main_df)
    log_df = pd.DataFrame({'data_classification': [data_classification_], 
                           'data_set': [main_dataset_name], 
                           'is_main': [1], 
                           'join_criteria': ['nan'],
                           'as_of_datetime': [file_TimeStamp],
                           'row_count': [row_count]
                          })

    if main_columns != r'*':
        #If the column list was specific and not a * then only return the columns that were asked for
        main_column_list = main_columns.replace('[','').replace(']','').replace(' ','').split(',')
        main_column_list = list(set(main_column_list))
        main_df = main_df[main_df.columns.intersection(main_column_list)]
    
    #df dataframe will be our holding dataframe that will then be joined and eventually returned to the calling app
    df = main_df
    
    #Get the list of datasets that we need to join into a dataframe
    #data_set_list_ = [x.lower() for x in data_set_list_]
    ds_df = pd.DataFrame(columns=['dataset_name','join_criteria'])
    for ds in data_set_list_:
        join_criteria = 'left'
        if ':' in ds:
            join_criteria = ds.split(':')[1]
            data_set_name = ds.split(':')[0]
            join_criteria = join_criteria.lower()
            data_set_name = data_set_name.lower()
        else:
            data_set_name = ds
        df2 = pd.DataFrame({'dataset_name': [data_set_name], 'join_criteria': [join_criteria]})
        ds_df = pd.concat([ds_df,df2], ignore_index=True)
    ds_df['dataset_name'] = ds_df['dataset_name'].str.lower()
    ds_df['join_criteria'] = ds_df['join_criteria'].str.lower()
    
    #Process each of the other data files in the same data classification and join if asked for
    for index, row in df_dl_wrapper_other.iterrows():
        dataset_name = row['DataSet']
        dataset_name = dataset_name.lower()
        dataset_name_df = ds_df[ds_df['dataset_name'] == dataset_name]
        if (len(dataset_name_df) > 0):
            dataset_name = dataset_name_df['dataset_name'].iloc[0]
            join_criteria = dataset_name_df['join_criteria'].iloc[0]
            #Get the parquet file location for the data that we need to join
            data_file = row['ParquetFileAndLocation']
            data_file = data_file.replace(r'{KDS_DL_ROOT}',kds_data_root)
            if (r'{KDS_DE_DATA}' in data_file):
                data_file = data_file.replace(r'{KDS_DE_DATA}',kds_data_de)
            if (r'{KDS_DS_DATA}' in data_file):
                data_file = data_file.replace(r'{KDS_DS_DATA}',kds_data_ds)
            #read the data parquet into a dataframe
            df_tmp = pd.read_parquet(data_file)
            #change the column names to lowercase
            df_tmp.columns = df_tmp.columns.str.lower()
            file_TimeStamp = datetime.datetime.fromtimestamp(os.stat(data_file).st_mtime).strftime("%m/%d/%Y %H:%M:%S")
            row_count = len(df_tmp)
            log_df = pd.concat([log_df,
                                pd.DataFrame({'data_classification': [data_classification_], 
                                              'data_set': [dataset_name], 
                                              'is_main': [0], 
                                              'join_criteria': [join_criteria],
                                              'as_of_datetime': [file_TimeStamp],
                                              'row_count': [row_count]
                                             })
                               ], ignore_index=True)

            #Get the columns that we need to return back to the calling app            
            columns = row['ReturnColumn'].lower()

            #Get the columns that we need to join on
            join_column = row['JoinColumn'].replace('[','').replace(']','').replace(' ','').lower().split(',')
            #for each of the columns that we have in the join if the column is of type string 
            #then lets convert it to lower case and create a new column that we will use for the join
            #after the join is done we will drop these columns before we send the data back to the app
            new_join_column_l = []
            new_join_column_r = []
            org_join_col = []
            for org_col in join_column:
                #Generate the new column name for the left and right side of the join
                new_col_l = org_col + '_4join_l'
                new_col_r = org_col + '_4join_r'
                #Store the original join column list so we can remove it from the right side of the dataframes before the merge
                #this avoids having column names with _x or _y in the result set.
                org_join_col += [org_col]
                #Add the new column name to the list
                new_join_column_l += [new_col_l]
                new_join_column_r += [new_col_r]
                #Create a new column in the main dataframe and the new dataframe
                #use try/catch to not break the code if the datatype of the column is not string
                #Can check the dtype since datetime also return dtype as object
                try:
                    df_tmp[new_col_r] = df_tmp[org_col].str.lower()
                except:
                    df_tmp[new_col_r] = df_tmp[org_col]

                try:
                    df[new_col_l] = df[org_col].str.lower()
                except:
                    df[new_col_l] = df[org_col]
                    

            if columns != r'*':
                column_list = columns.replace('[','').replace(']','').replace(' ','').split(',')
                column_list = column_list + new_join_column_r
                column_list = list(set(column_list))
                df_tmp = df_tmp[df_tmp.columns.intersection(column_list)]

            #Remove the join columns from the right side dataframe so we don't get the _x or _y columns
            for col in org_join_col:
                if col in df_tmp.columns:
                    del df_tmp[col]
            
            #Now merge the 2 dataframes
            df = df.merge(df_tmp, left_on=new_join_column_l, right_on=new_join_column_r, how=join_criteria, copy=True)
            
            #Now drop the join columns that we used so as to return the original column list back to the calling app
            remove_join_columns = new_join_column_l + new_join_column_r
            df = df.drop(columns=remove_join_columns)
            
    return df, log_df
    
def filter_data(df, data_filter):
    lst_filter_ = data_filter.split(':')
    lst_col=[]
    for f in lst_filter_:
        lst_f=f.split('=')
        col_name = lst_f[0][0:].lower()
        lst_col=lst_col + [col_name]
        values_ = lst_f[1][0:]
        lst_values_ = values_.split(',')
        df=df[df[col_name].isin(lst_values_)]
    
    return df
    
def get_info(): return pd.read_json(kds_data_wrapper)
    
def get_dl_data(**kwargs):
    """
    joins the various requested datasets and also filters them if needed.
    

    Parameters
    ----------
    **kwargs : 
        data_classification: str, required
            This is request what collection/class of data is being requsted.
            eg: data_classification = 'project'
        data_set_list: list[str], optional
            This is an optional parameter containing names of various datasets that need to be joined and how to join them
            There can be 2 parts to this string seperated by a ':' (optional)
            if no ':' is supplied then the string is the dataset name and the type of join is 'left'
            if ':' is supplied then the left side of the string is the dataset name and right is how to join
            eg: data_set_lst = ['climate','population']
        data_used_for: str, required
            This is a string that is passed in to let the call know what this data is going to be used for.
        data_filter: str, optional
            This is an optional paramter that specifies the filters to be applied. This string is a name value pair seperated by 
            '='. If more than 1 filter need to be applied and then multiples can be passed by separating them by ':'
            eg: data_filter = 'PROJECT_Number=103066,103057,101417:INDUSTRY=Power'
    Raises
    ------
    ValueError
        If required parameters were not supplied value error is raised.

    Returns
    -------
    ret_df_ : pandas.DataFrame
        dataframe containing joined and filtered data

    """

    #Get the calling script name
    frame_ = inspect.stack()[1]
    module_ = inspect.getmodule(frame_[0])
    try:
        filename_ = module_.__file__
    except:
        filename_ = 'no script probably jupiter notebook'
    computer_name_ = socket.gethostname()
    
    data_classification_ = None
    data_set_list_ = ''
    data_used_for_ = None
    data_filter_ = None
    ret_df_ = pd.DataFrame()
    user_ = os.getlogin()
    curr_date_ = datetime.datetime.now()
    
    if kwargs.get('data_classification') is not None:
        data_classification_ = kwargs.get('data_classification')
    if kwargs.get('data_set_list') is not None:
        data_set_list_ = kwargs.get('data_set_list')
    if kwargs.get('data_used_for') is not None:
        data_used_for_ = kwargs.get('data_used_for')
    if kwargs.get('data_filter') is not None:
        data_filter_ = kwargs.get('data_filter')
    
    if data_classification_ is None:
        raise ValueError("data_classification parameter wasn't passed to the call")
    if (data_used_for_ is None) or (len(data_used_for_) == 0):
        raise ValueError("data_used_for parameter wasn't passed to the call")
    
    logutilobj.log_info(f"data_classification={data_classification_}~~data_set_list={data_set_list_}~~user={user_}~~exec_dt={curr_date_}")
    
    ret_df_,log_df = get_data(data_classification_, data_set_list_)
    
    if data_filter_ is not None:
        ret_df_ = filter_data(ret_df_, data_filter_)
        
    log_df['user'] = user_
    log_df['dataset_list'] = ','.join(data_set_list_)
    log_df['data_used_for'] = data_used_for_
    log_df['data_filter'] = data_filter_
    log_df['data_request_datetime'] = curr_date_.strftime("%m/%d/%Y %H:%M:%S")
    log_df['calling_script'] = filename_
    log_df['calling_computer'] = computer_name_
    log_file_name = kds_data_wrapper_log_dir + curr_date_.strftime("%Y%m%d_%H%M%S_%f_") + f'{random.randrange(1, 10**3):03}.parquet'
    log_df.to_parquet(log_file_name)
    
    return ret_df_

#################################################################
# Main function equivalent. 
# Only gets executed when this file is run independently  
#################################################################
if __name__ == '__main__':          
    
   _filename = sys.argv[0].split(os.path.sep)[-1].split('.')[0]
   print(_filename)
   
   try:
       #Invoke the log_info class method for the logutilobj.
       logutilobj.log_info('Main function test begin.')
       data_classification = 'project'
       data_set_lst = ['climate:inner','population']
       data_used_for = 'HG_Testing_1.0'
       #data_filter = 'PROJECT_Number=103066,103057,101417:INDUSTRY=Power' #data_filter=data_filter, 
       x = get_dl_data(data_classification=data_classification, data_set_list=data_set_lst, data_used_for=data_used_for)
    
   except:
       err = sys.exc_info()[0]
       errStackTrace = traceback.format_exc(limit=None, chain=True)
       #Invoke the log_error class method for the logutilobj.
       logutilobj.log_error(errStackTrace)
       logutilobj.delete() 
       print('completed the except section')



from setuptools import setup, find_packages
#from distutils.core import setup

with open("README.md", "r") as fh:
    long_description = fh.read()
    
# Setting up
setup(
       # the name must match the folder name 'kdslib'
        name="kdslib", 
        version='1.0.7',
        author="kdslibraries",
        author_email="kdslibraries@aol.com",
        description='KDS logging and data connection helpers',
        long_description=long_description,
        long_description_content_type="text/markdown",
        packages=find_packages(),
        install_requires=['msal','requests','sqlalchemy','teradata','teradatasql','pymssql','teradatasqlalchemy'], # add any additional packages that 
        # needs to be installed along with your package. Eg: 'msal'
        
        keywords=['python', 'kdslib'],
        classifiers= [
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)